import rule34
import pytest
import os


def test_Sync():
    r34 = rule34.Sync()
    assert r34.l is not None
    assert r34.getImages("gay") is not None


def test_async():
    import asyncio
    loop = asyncio.get_event_loop()
    r34 = rule34.Rule34(loop)
    assert (len(loop.run_until_complete(r34.getImages("gay"))) == 100)


def test_postData():
    r34 = rule34.Sync()
    assert (r34.getPostData(1) is not None)


def test_URLGen():
    r34 = rule34.Sync()
    expectedString = "https://rule34.xxx/index.php?page=dapi&s=post&q=index&limit=50&tags=gay&deleted=show&rating:explicit"
    assert (r34.URLGen(tags="gay", limit=50, deleted=True) == expectedString)


def test_URLGenPIDLimit():
    r34 = rule34.Sync()
    try:
        r34.URLGen(tags="gay", PID=2001)
    except rule34.Request_Rejected:
        pass
    except Exception as e:
        raise rule34.SelfTest_Failed(e)


def test_URLGenReturnNone():
    r34 = rule34.Sync()
    assert r34.URLGen() is None


def test_Download():
    r34 = rule34.Sync()
    downloadURL = 'https://img.rule34.xxx/images/2003/b90ae3f67eaa30669939531292d90e55d58325af.jpg'
    name = r34.download(downloadURL)
    assert name is not None
    assert os.path.isfile(name)
    assert os.path.getsize(name) >= 76600
    os.unlink(name)


def test_DonwloadNameHandler():
    r34 = rule34.Sync()
    downloadURL = 'https://img.rule34.xxx/images/2003/b90ae3f67eaa30669939531292d90e55d58325af.jpg'
    names = []
    names.append(r34.download(downloadURL))
    names.append(r34.download(downloadURL))
    for name in names:
        assert name is not None
        assert os.path.isfile(name)
        assert os.path.getsize(name) >= 76600
        os.unlink(name)


def test_DownloadErrorCatch():
    r34 = rule34.Sync()
    downloadURL = 'https://img.rule34.xxx/borris/2003/b90ae3f67eaa30669939531292d90e55d58325af.jpg'
    try:
        name = r34.download(downloadURL)
        os.unlink(name)
    except rule34.Rule34_Error:
        pass
    except Exception as e:
        raise rule34.SelfTest_Failed(e)


def test_TotalImages():
    r34 = rule34.Sync()
    assert r34.totalImages("gay") > 1000

def test_imageGatherNew():
    r34 = rule34.Sync()
    assert (r34.getImages("straight") is not None)

def test_imageGatherNew_total():
    r34 = rule34.Sync()
    assert (len(r34.getImages("straight")) == 100)

def test_imageGatherNew_OverridePID():
    r34 = rule34.Sync()
    assert (r34.getImages("straight",
                             singlePage=True,
                             OverridePID=1) is not None)

def test_imageGatherNew_fuzzy():
    r34 = rule34.Sync()
    assert (r34.getImages("vore",
                             singlePage=True,
                             fuzzy=True) is not None)


def test_imageGatherNew_Contradiction():
    r34 = rule34.Sync()
    assert (r34.getImages("vore",
                             singlePage=False,
                             randomPID=True) is not None)


def test_imageGatherNew_nonExist():
    r34 = rule34.Sync()
    assert (r34.getImages("DNATESTMAGICCOODENOTHINGWILLRETURN") is None)


def test_imageGatherNew_rejectRequest():
    r34 = rule34.Sync()
    try:
        r34.getImages("straight", fuzzy=True, OverridePID=2001)
    except rule34.Request_Rejected:
        pass
    except Exception as e:
        raise rule34.SelfTest_Failed(e)



"""
Hi there, LordOfPolls here. I almost guarantee this isnt how an __init__.py is supposed to be written, but *shrugs*, it works
"""

from .objectClasses import Rule34Post  # A class representing a rule34 post

# Both rule34 api run modes
from .rule34 import Rule34  # Async
from .rule34 import Sync  # Sync

# Error Types
from .rule34 import Request_Rejected
from .rule34 import Rule34_Error
from .rule34 import SelfTest_Failed


from __future__ import print_function

import asyncio
import math
import random
import os
from collections import defaultdict
from xml.etree import cElementTree as ET
import aiohttp
import async_timeout
import warnings

from .objectClasses import Rule34Post

class Rule34_Error(Exception):  # pragma: no cover
    """Rule34 rejected you"""
    def __init__(self, message, *args):
        self.message = message
        super(Rule34_Error, self).__init__(message, *args)


class Request_Rejected(Exception):  # pragma: no cover
    """The Rule34 API wrapper rejected your request"""
    def __init__(self, message, *args):
        self.message = message
        super(Request_Rejected, self).__init__(message, *args)


class SelfTest_Failed(Exception):  # pragma: no cover
    """The self test failed"""
    def __init__(self, message, *args):
        self.message = message
        super(SelfTest_Failed, self).__init__(message, *args)


class Rule34:
    def __init__(self, loop, timeout=10):
        """
        :param loop: the event loop
        :param timeout: how long requests are allowed to run until timing out
        """
        self.session = aiohttp.ClientSession(loop=loop)
        self.timeout = timeout
        self.loop = loop

    def session(self):
        return self.session

    def ParseXML(self, rawXML):
        """Parses entities as well as attributes following this XML-to-JSON "specification"
            Using https://stackoverflow.com/a/10077069"""
        if "Search error: API limited due to abuse" in str(rawXML.items()):
            raise Rule34_Error('Rule34 rejected your request due to "API abuse"')

        d = {rawXML.tag: {} if rawXML.attrib else None}
        children = list(rawXML)
        if children:
            dd = defaultdict(list)
            for dc in map(self.ParseXML, children):
                for k, v in dc.items():
                    dd[k].append(v)
            d = {rawXML.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
        if rawXML.attrib:
            d[rawXML.tag].update(('@' + k, v) for k, v in rawXML.attrib.items())
        if rawXML.text:
            text = rawXML.text.strip()
            if children or rawXML.attrib:
                if text:
                    d[rawXML.tag]['#text'] = text
            else:
                d[rawXML.tag] = text
        return d

    @staticmethod
    def urlGen(tags=None, limit=None, ID=None, PID=None, deleted=None, rating=None, **kwargs):
        """Generates a URL to access the api using your input:
        :param tags: str ||The tags to search for. Any tag combination that works on the web site will work here. This includes all the meta-tags
        :param limit: str ||How many posts you want to retrieve
        :param ID: int ||The post id.
        :param PID: int ||The page number.
        :param deleted: bool||If True, deleted posts will be included in the data
        :param rating: The rating of the images, defaults to explicit
        :param kwargs:
        :return: url string, or None

        All arguments that accept strings *can* accept int, but strings are recommended
        If none of these arguments are passed, None will be returned
        """
        # I have no intentions of adding "&last_id=" simply because its response can easily be massive, and all it returns is ``<post deleted="[ID]" md5="[String]"/>`` which has no use as far as im aware
        URL = "https://rule34.xxx/index.php?page=dapi&s=post&q=index"
        if PID is not None:
            if PID > 2000:
                raise Request_Rejected("Rule34 will reject PIDs over 2000")
            URL += "&pid={}".format(PID)
        if limit is not None:
            URL += "&limit={}".format(limit)
        if ID is not None:
            URL += "&id={}".format(ID)
        if tags is not None:
            tags = str(tags).replace(" ", "+")
            URL += "&tags={}".format(tags)
        if deleted:
            URL += "&deleted=show"
        if PID is not None or limit is not None or ID is not None or tags is not None:
            if ID is not None:
                return URL
            else:
                if rating:
                    return URL + f"&rating:{rating}"
                else:
                    return URL + f"&rating:explicit"
        else:
            return None

    async def totalImages(self, tags):
        """Returns the total amount of images for the tag
        :param tags:
        :return: int
        """
        if self.session.closed:
            self.session = aiohttp.ClientSession(loop=self.loop)
        with async_timeout.timeout(10):
            url = self.urlGen(tags=tags, PID=0)
            async with self.session.get(url=url) as rawXML:
                rawXML = await rawXML.read()
                rawXML = ET.XML(rawXML)
                XML = self.ParseXML(rawXML)
            await self.session.close()
            return int(XML['posts']['@count'])
        return None

    async def getImages(self, tags, fuzzy=False, singlePage=True, randomPID=True, OverridePID=None, rating=None):
        """gatherers a list of image's and their respective data -- replacing getImageURLS
        :param tags: the tags you're searching
        :param fuzzy: enable or disable fuzzy search, default disabled
        :param singlePage: when enabled, limits the search to one page (100 images), default disabled
        :param randomPID: when enabled, a random pageID is used, if singlePage is disabled, this is disabled
        :param OverridePID: Allows you to specify a PID
        :param rating: The rating of the images, defaults to explicit
        :return: list
        """
        if self.session.closed:  # Verify we have an active session (avoids errors)
            self.session = aiohttp.ClientSession(loop=self.loop)
        if fuzzy:
            tags = tags.split(" ")
            for tag in tags:
                tag = tag + "~"
            temp = " "
            tags = temp.join(tags)
        if randomPID is True and singlePage is False:
            randomPID = False
        num = await self.totalImages(tags)

        if num != 0:
            if OverridePID is not None:
                if OverridePID >2000:
                    raise Request_Rejected("Rule34 will reject PIDs over 2000")
                PID = OverridePID
            elif randomPID:
                maxPID = 2000
                if math.floor(num/100) < maxPID:
                    maxPID = math.floor(num/100)
                PID = random.randint(0, maxPID)
            else:
                PID = 0
            imgList = []
            XML = None
            t = True
            while t:
                tempURL = self.urlGen(tags=tags, PID=PID, rating=rating)
                with async_timeout.timeout(self.timeout):
                    if self.session.closed:
                        self.session = aiohttp.ClientSession(loop=self.loop)
                    async with self.session.get(url=tempURL) as XML:
                        XML = await XML.read()
                        XML = ET.XML(XML)
                        XML = self.ParseXML(XML)
                if XML is None:
                    return None
                if len(imgList) >= int(XML['posts']['@count']):  # "if we're out of images to process"
                    t = False  # "end the loop"
                else:
                    # if isinstance(XML['posts']['post'], dict):
                    #     # Data is in dict, aka one image
                    #     image = Rule34Post()  # create an image object
                    #     image.parse(XML['posts']['post'])  # parse data into object
                    #     imgList.append(image)  # add to image list
                    # else:
                    #     # Data is string, aka many images
                    for post in XML['posts']['post']:
                        image = Rule34Post()
                        image.parse(post)
                        imgList.append(image)
                if singlePage:
                    await self.session.close()
                    return imgList
                PID += 1
            await self.session.close()
            return imgList
        else:
            await self.session.close()
            return None

    async def getPostData(self, PostID):
        """Returns a dict with all the information available about the post
        :param PostID: The ID of the post
        :return: dict
        """

        if self.session.closed:
            self.session = aiohttp.ClientSession(loop=self.loop)
        url = self.urlGen(ID=str(PostID))
        XML =None
        with async_timeout.timeout(10):
            async with self.session.get(url=url) as XML:
                XML = await XML.read()
            await self.session.close()
            try:
                XML = self.ParseXML(ET.XML(XML))
                data = XML['posts']['post']
            except ValueError:
                return None
            return data
        return None


    async def download(self, URL, destination=None):
        try:
            if self.session.closed:
                self.session = aiohttp.ClientSession(loop=self.loop)
            async with self.session.get(URL) as resp:
                assert resp.status == 200
                i = 7
                name = URL.split("/")[-1][-7:]
                if destination:
                    name = f"{destination}/{name}"
                while os.path.isfile(name):
                    i += 1
                    name = URL.split("/")[-1][-i:]
                data = await resp.read()
            await self.session.close()
            with open(name, "wb") as f:
                f.write(data)
        except Exception as e:
            raise Rule34_Error(e)
            return None
        return name


class Sync:
    """Allows you to run the module without worrying about async"""
    def __init__(self):
        self.l = asyncio.get_event_loop()
        self.r = Rule34(self.l)

    def sessionClose(self):  # pragma: no cover
        if not self.r.session.closed:
            self.l.run_until_complete(self.r.session.close())

    def getImages(self, tags, fuzzy=False, singlePage=True, randomPID=True, OverridePID=None, rating=None):
        """gatherers a list of image's and their respective data -- replacing getImageURLS
        :param tags: the tags you're searching
        :param fuzzy: enable or disable fuzzy search, default disabled
        :param singlePage: when enabled, limits the search to one page (100 images), default disabled
        :param randomPID: when enabled, a random pageID is used, if singlePage is disabled, this is disabled
        :param rating: The rating of the images, defaults to explicit
        :param OverridePID: Allows you to specify a PID
        :return: list
        """
        data = self.l.run_until_complete(self.r.getImages(tags, fuzzy, singlePage, randomPID, OverridePID, rating))
        return data

    def getPostData(self, tags):
        """Returns a dict with all the information available about the post
        :param PostID: The ID of the post
        :return: dict
        """
        return self.l.run_until_complete(self.r.getPostData(tags))

    def totalImages(self, tags):
        """Returns the total amount of images for the tag
        :param tags:
        :return: int
        """
        return self.l.run_until_complete(self.r.totalImages(tags))

    @staticmethod
    def URLGen(tags=None, limit=None, id=None, PID=None, deleted=None, rating=None):
        """Generates a URL to access the api using your input:
        :param tags: str ||The tags to search for. Any tag combination that works on the web site will work here. This includes all the meta-tags
        :param limit: str ||How many posts you want to retrieve
        :param id: int ||The post id.
        :param PID: int ||The page number.
        :param deleted: bool||If True, deleted posts will be included in the data
        :param kwargs:
        :return: url string, or None
        All arguments that accept strings *can* accept int, but strings are recommended
        If none of these arguments are passed, None will be returned
        """
        return Rule34.urlGen(tags, limit, id, PID, deleted, rating)

    def download(self, url):
        return self.l.run_until_complete(self.r.download(url))



from setuptools import setup, find_packages

setup(
    name='rule34',
    version="1.7.0",
    description='An async API wrapper for rule34.xxx',
    long_description= 'An async API wrapper for rule34.xxx using aiohttp',
    long_description_content_type='text/markdown',
    url='https://github.com/LordOfPolls/Rule34-API-Wrapper',
    author='LordOfPolls',
    author_email='ddavidallen13@gmail.com',
    license='MIT',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
         'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.

        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'],
    keywords='rule34, porn, api, wrapper',
    project_urls={'Source': 'https://github.com/LordOfPolls/Rule34-API-Wrapper'},
    packages=find_packages(include=['rule34']),
    install_requires=['asyncio', 'aiohttp'],
    python_requires='>=3.5',
    py_modules=["rule34"]
)
